@@ -1,22 +1,20 @@
 /* Directory hashing for GNU Make.
-Copyright (C) 1988, 1989, 1991, 1992, 1993, 1994, 1995, 1996, 1997,
-2002 Free Software Foundation, Inc.
+Copyright (C) 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996, 1997,
+1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006 Free Software
+Foundation, Inc.
 This file is part of GNU Make.
 
-GNU Make is free software; you can redistribute it and/or modify
-it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
-any later version.
+GNU Make is free software; you can redistribute it and/or modify it under the
+terms of the GNU General Public License as published by the Free Software
+Foundation; either version 2, or (at your option) any later version.
 
-GNU Make is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-GNU General Public License for more details.
+GNU Make is distributed in the hope that it will be useful, but WITHOUT ANY
+WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
+A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public License
-along with GNU Make; see the file COPYING.  If not, write to
-the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
-Boston, MA 02111-1307, USA.  */
+You should have received a copy of the GNU General Public License along with
+GNU Make; see the file COPYING.  If not, write to the Free Software
+Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.  */
 
 #include "make.h"
 #include "hash.h"
@@ -70,8 +68,7 @@ extern char *vmsify PARAMS ((char *name, int type));
 #endif
 
 static char *
-dosify (filename)
-     char *filename;
+dosify (char *filename)
 {
   static char dos_filename[14];
   char *df;
@@ -121,14 +118,9 @@ dosify (filename)
 
 #ifdef HAVE_CASE_INSENSITIVE_FS
 static char *
-downcase (filename)
-     char *filename;
+downcase (char *filename)
 {
-#ifdef _AMIGA
-  static char new_filename[136];
-#else
-  static char new_filename[PATH_MAX];
-#endif
+  static PATH_VAR (new_filename);
   char *df;
   int i;
 
@@ -153,8 +145,7 @@ downcase (filename)
 #ifdef VMS
 
 static int
-vms_hash (name)
-    char *name;
+vms_hash (char *name)
 {
   int h = 0;
   int g;
@@ -162,7 +153,11 @@ vms_hash (name)
   while (*name)
     {
       unsigned char uc = *name;
+#ifdef HAVE_CASE_INSENSITIVE_FS
       h = (h << 4) + (isupper (uc) ? tolower (uc) : uc);
+#else
+      h = (h << 4) + uc;
+#endif
       name++;
       g = h & 0xf0000000;
       if (g)
@@ -176,9 +171,7 @@ vms_hash (name)
 
 /* fake stat entry for a directory */
 static int
-vmsstat_dir (name, st)
-    char *name;
-    struct stat *st;
+vmsstat_dir (char *name, struct stat *st)
 {
   char *s;
   int h;
@@ -248,13 +241,13 @@ struct directory_contents
   };
 
 static unsigned long
-directory_contents_hash_1 (key_0)
-    const void *key_0;
+directory_contents_hash_1 (const void *key_0)
 {
   struct directory_contents const *key = (struct directory_contents const *) key_0;
   unsigned long hash;
 
 #ifdef WINDOWS32
+  hash = 0;
   ISTRING_HASH_1 (key->path_key, hash);
   hash ^= ((unsigned int) key->dev << 4) ^ (unsigned int) key->ctime;
 #else
@@ -271,13 +264,13 @@ directory_contents_hash_1 (key_0)
 }
 
 static unsigned long
-directory_contents_hash_2 (key_0)
-    const void *key_0;
+directory_contents_hash_2 (const void *key_0)
 {
   struct directory_contents const *key = (struct directory_contents const *) key_0;
   unsigned long hash;
 
 #ifdef WINDOWS32
+  hash = 0;
   ISTRING_HASH_2 (key->path_key, hash);
   hash ^= ((unsigned int) key->dev << 4) ^ (unsigned int) ~key->ctime;
 #else
@@ -294,10 +287,19 @@ directory_contents_hash_2 (key_0)
   return hash;
 }
 
+/* Sometimes it's OK to use subtraction to get this value:
+     result = X - Y;
+   But, if we're not sure of the type of X and Y they may be too large for an
+   int (on a 64-bit system for example).  So, use ?: instead.
+   See Savannah bug #15534.
+
+   NOTE!  This macro has side-effects!
+*/
+
+#define MAKECMP(_x,_y)  ((_x)<(_y)?-1:((_x)==(_y)?0:1))
+
 static int
-directory_contents_hash_cmp (xv, yv)
-    const void *xv;
-    const void *yv;
+directory_contents_hash_cmp (const void *xv, const void *yv)
 {
   struct directory_contents const *x = (struct directory_contents const *) xv;
   struct directory_contents const *y = (struct directory_contents const *) yv;
@@ -307,28 +309,28 @@ directory_contents_hash_cmp (xv, yv)
   ISTRING_COMPARE (x->path_key, y->path_key, result);
   if (result)
     return result;
-  result = x->ctime - y->ctime;
+  result = MAKECMP(x->ctime, y->ctime);
   if (result)
     return result;
 #else
 # ifdef VMS
-  result = x->ino[0] - y->ino[0];
+  result = MAKECMP(x->ino[0], y->ino[0]);
   if (result)
     return result;
-  result = x->ino[1] - y->ino[1];
+  result = MAKECMP(x->ino[1], y->ino[1]);
   if (result)
     return result;
-  result = x->ino[2] - y->ino[2];
+  result = MAKECMP(x->ino[2], y->ino[2]);
   if (result)
     return result;
 # else
-  result = x->ino - y->ino;
+  result = MAKECMP(x->ino, y->ino);
   if (result)
     return result;
 # endif
 #endif /* WINDOWS32 */
 
-  return x->dev - y->dev;
+  return MAKECMP(x->dev, y->dev);
 }
 
 /* Table of directory contents hashed by device and inode number.  */
@@ -345,23 +347,19 @@ struct directory
   };
 
 static unsigned long
-directory_hash_1 (key)
-    const void *key;
+directory_hash_1 (const void *key)
 {
   return_ISTRING_HASH_1 (((struct directory const *) key)->name);
 }
 
 static unsigned long
-directory_hash_2 (key)
-    const void *key;
+directory_hash_2 (const void *key)
 {
   return_ISTRING_HASH_2 (((struct directory const *) key)->name);
 }
 
 static int
-directory_hash_cmp (x, y)
-    const void *x;
-    const void *y;
+directory_hash_cmp (const void *x, const void *y)
 {
   return_ISTRING_COMPARE (((struct directory const *) x)->name,
 			  ((struct directory const *) y)->name);
@@ -387,23 +385,19 @@ struct dirfile
   };
 
 static unsigned long
-dirfile_hash_1 (key)
-    const void *key;
+dirfile_hash_1 (const void *key)
 {
   return_ISTRING_HASH_1 (((struct dirfile const *) key)->name);
 }
 
 static unsigned long
-dirfile_hash_2 (key)
-    const void *key;
+dirfile_hash_2 (const void *key)
 {
   return_ISTRING_HASH_2 (((struct dirfile const *) key)->name);
 }
 
 static int
-dirfile_hash_cmp (xv, yv)
-    const void *xv;
-    const void *yv;
+dirfile_hash_cmp (const void *xv, const void *yv)
 {
   struct dirfile const *x = ((struct dirfile const *) xv);
   struct dirfile const *y = ((struct dirfile const *) yv);
@@ -423,8 +417,7 @@ static struct directory *find_directory PARAMS ((char *name));
 /* Find the directory named NAME and return its `struct directory'.  */
 
 static struct directory *
-find_directory (name)
-     register char *name;
+find_directory (char *name)
 {
   register char *p;
   register struct directory *dir;
@@ -435,9 +428,9 @@ find_directory (name)
   char* w32_path;
   char  fs_label[BUFSIZ];
   char  fs_type[BUFSIZ];
-  long  fs_serno;
-  long  fs_flags;
-  long  fs_len;
+  unsigned long  fs_serno;
+  unsigned long  fs_flags;
+  unsigned long  fs_len;
 #endif
 #ifdef VMS
   if ((*name == '.') && (*(name+1) == 0))
@@ -473,7 +466,7 @@ find_directory (name)
 #ifdef VMS
       r = vmsstat_dir (name, &st);
 #else
-      r = stat (name, &st);
+      EINTRLOOP (r, stat (name, &st));
 #endif
 
 #ifdef WINDOWS32
@@ -554,7 +547,7 @@ find_directory (name)
 # endif
 #endif /* WINDOWS32 */
 	      hash_insert_at (&directory_contents, dc, dc_slot);
-	      dc->dirstream = opendir (name);
+	      ENULLLOOP (dc->dirstream, opendir (name));
 	      if (dc->dirstream == 0)
                 /* Couldn't open the directory.  Mark this by
                    setting the `files' member to a nil pointer.  */
@@ -584,9 +577,7 @@ find_directory (name)
    FILENAME must contain no slashes.  */
 
 static int
-dir_contents_file_exists_p (dir, filename)
-     register struct directory_contents *dir;
-     register char *filename;
+dir_contents_file_exists_p (struct directory_contents *dir, char *filename)
 {
   unsigned int hash;
   struct dirfile *df;
@@ -609,6 +600,11 @@ dir_contents_file_exists_p (dir, filename)
   filename = downcase (filename);
 #endif
 
+#ifdef __EMX__
+  if (filename != 0)
+    _fnlwr (filename); /* lower case for FAT drives */
+#endif
+
 #ifdef VMS
   filename = vmsify (filename,0);
 #endif
@@ -643,21 +639,28 @@ dir_contents_file_exists_p (dir, filename)
        * filesystems force a rehash always as mtime does not change
        * on directories (ugh!).
        */
-      if (dir->path_key
-	  && (dir->fs_flags & FS_FAT
-	      || (stat(dir->path_key, &st) == 0
-		  && st.st_mtime > dir->mtime)))
+      if (dir->path_key)
 	{
-	  /* reset date stamp to show most recent re-process */
-	  dir->mtime = st.st_mtime;
+          if ((dir->fs_flags & FS_FAT) != 0)
+	    {
+	      dir->mtime = time ((time_t *) 0);
+	      rehash = 1;
+	    }
+	  else if (stat(dir->path_key, &st) == 0 && st.st_mtime > dir->mtime)
+	    {
+	      /* reset date stamp to show most recent re-process.  */
+	      dir->mtime = st.st_mtime;
+	      rehash = 1;
+	    }
 
-	  /* make sure directory can still be opened */
-	  dir->dirstream = opendir(dir->path_key);
+          /* If it has been already read in, all done.  */
+	  if (!rehash)
+	    return 0;
 
-	  if (dir->dirstream)
-	    rehash = 1;
-	  else
-	    return 0; /* couldn't re-read - fail */
+          /* make sure directory can still be opened; if not return.  */
+          dir->dirstream = opendir(dir->path_key);
+          if (!dir->dirstream)
+            return 0;
 	}
       else
 #endif
@@ -665,13 +668,17 @@ dir_contents_file_exists_p (dir, filename)
 	return 0;
     }
 
-  while ((d = readdir (dir->dirstream)) != 0)
+  while (1)
     {
       /* Enter the file in the hash table.  */
       unsigned int len;
       struct dirfile dirfile_key;
       struct dirfile **dirfile_slot;
 
+      ENULLLOOP (d, readdir (dir->dirstream));
+      if (d == 0)
+        break;
+
 #if defined(VMS) && defined(HAVE_DIRENT_H)
       /* In VMS we get file versions too, which have to be stripped off */
       {
@@ -724,9 +731,7 @@ dir_contents_file_exists_p (dir, filename)
    FILENAME must contain no slashes.  */
 
 int
-dir_file_exists_p (dirname, filename)
-     register char *dirname;
-     register char *filename;
+dir_file_exists_p (char *dirname, char *filename)
 {
   return dir_contents_file_exists_p (find_directory (dirname)->contents,
 				     filename);
@@ -735,8 +740,7 @@ dir_file_exists_p (dirname, filename)
 /* Return 1 if the file named NAME exists.  */
 
 int
-file_exists_p (name)
-     register char *name;
+file_exists_p (char *name)
 {
   char *dirend;
   char *dirname;
@@ -751,8 +755,7 @@ file_exists_p (name)
   dirend = strrchr (name, ']');
   if (dirend == 0)
     dirend = strrchr (name, ':');
-  dirend++;
-  if (dirend == (char *)1)
+  if (dirend == (char *)0)
     return dir_file_exists_p ("[]", name);
 #else /* !VMS */
   dirend = strrchr (name, '/');
@@ -798,8 +801,7 @@ file_exists_p (name)
    as an intermediate file, and it has failed.  */
 
 void
-file_impossible (filename)
-     register char *filename;
+file_impossible (char *filename)
 {
   char *dirend;
   register char *p = filename;
@@ -882,8 +884,7 @@ file_impossible (filename)
 /* Return nonzero if FILENAME has been marked impossible.  */
 
 int
-file_impossible_p (filename)
-     char *filename;
+file_impossible_p (char *filename)
 {
   char *dirend;
   register char *p = filename;
@@ -964,8 +965,7 @@ file_impossible_p (filename)
    directory hash table that matches DIR.  */
 
 char *
-dir_name (dir)
-     char *dir;
+dir_name (char *dir)
 {
   return find_directory (dir)->name;
 }
@@ -973,7 +973,7 @@ dir_name (dir)
 /* Print the data base of directories.  */
 
 void
-print_dir_data_base ()
+print_dir_data_base (void)
 {
   register unsigned int files;
   register unsigned int impossible;
@@ -1096,8 +1096,7 @@ static __ptr_t open_dirstream PARAMS ((const char *));
 static struct dirent *read_dirstream PARAMS ((__ptr_t));
 
 static __ptr_t
-open_dirstream (directory)
-     const char *directory;
+open_dirstream (const char *directory)
 {
   struct dirstream *new;
   struct directory *dir = find_directory ((char *)directory);
@@ -1120,8 +1119,7 @@ open_dirstream (directory)
 }
 
 static struct dirent *
-read_dirstream (stream)
-     __ptr_t stream;
+read_dirstream (__ptr_t stream)
 {
   struct dirstream *const ds = (struct dirstream *) stream;
   struct directory_contents *dc = ds->contents;
@@ -1148,6 +1146,12 @@ read_dirstream (stream)
 	      buf = xmalloc (bufsz);
 	    }
 	  d = (struct dirent *) buf;
+#ifdef __MINGW32__
+# if __MINGW32_MAJOR_VERSION < 3 || (__MINGW32_MAJOR_VERSION == 3 && \
+				     __MINGW32_MINOR_VERSION == 0)
+	  d->d_name = xmalloc(len);
+# endif
+#endif
 	  FAKE_DIR_ENTRY (d);
 #ifdef _DIRENT_HAVE_D_NAMLEN
 	  d->d_namlen = len - 1;
@@ -1164,11 +1168,10 @@ read_dirstream (stream)
 }
 
 static void
-ansi_free(p)
-  void *p;
+ansi_free (void *p)
 {
-    if (p)
-      free(p);
+  if (p)
+    free(p);
 }
 
 /* On 64 bit ReliantUNIX (5.44 and above) in LFS mode, stat() is actually a
@@ -1177,21 +1180,22 @@ ansi_free(p)
  */
 #ifndef stat
 # ifndef VMS
-extern int stat ();
+extern int stat PARAMS ((const char *path, struct stat *sbuf));
 # endif
 # define local_stat stat
 #else
-static int local_stat (path, buf)
-    char *path;
-    struct stat *buf;
+static int
+local_stat (const char *path, struct stat *buf)
 {
-  return stat (path, buf);
+  int e;
+
+  EINTRLOOP (e, stat (path, buf));
+  return e;
 }
 #endif
 
 void
-dir_setup_glob (gl)
-     glob_t *gl;
+dir_setup_glob (glob_t *gl)
 {
   /* Bogus sunos4 compiler complains (!) about & before functions.  */
   gl->gl_opendir = open_dirstream;
@@ -1203,7 +1207,7 @@ dir_setup_glob (gl)
 }
 
 void
-hash_init_directories ()
+hash_init_directories (void)
 {
   hash_init (&directories, DIRECTORY_BUCKETS,
 	     directory_hash_1, directory_hash_2, directory_hash_cmp);
